# Custom expanded version of SHA512
import hashlib


class PMHash:
    digest_size = 256
    name = 'PMHash'

    def __init__(self, b=b""):
        self.b = b

    def update(self, b):
        self.b += b

    def digest(self):
        return hashlib.sha512(self.b + b'\0').digest() + hashlib.sha512(self.b + b'\1').digest() + hashlib.sha512(
            self.b + b'\2').digest() + hashlib.sha512(self.b + b'\3').digest()

    def hexdigest(self):
        return self.digest().hex()

    def copy(self):
        return PMHash(self.b)


def pmhash(b=b""):
    return PMHash(b)



import base64
import bcrypt
import os

PM_VERSION = 4


def bcrypt_b64_encode(s):  # The joy of bcrypt
    bcrypt_base64 = b"./ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
    std_base64chars = b"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
    s = base64.b64encode(s)
    return s.translate(bytes.maketrans(std_base64chars, bcrypt_base64))


def hash_password_3(hash_class, password, salt, modulus):
    salt = (salt + b"proton")[:16]
    salt = bcrypt_b64_encode(salt)[:22]
    hashed = bcrypt.hashpw(password, b"$2y$10$" + salt)
    return hash_class(hashed + modulus).digest()


def hash_password(hash_class, password, salt, modulus, version):
    if version == 4 or version == 3:
        return hash_password_3(hash_class, password, salt, modulus)

    raise ValueError('Unsupported auth version')


def long_length(n):
    return (n.bit_length() + 7) // 8


def bytes_to_long(s):
    return int.from_bytes(s, 'little')


def long_to_bytes(n):
    return n.to_bytes(long_length(n), 'little')


def get_random(nbytes):
    return bytes_to_long(os.urandom(nbytes))


def get_random_of_length(nbytes):
    offset = (nbytes * 8) - 1
    return get_random(nbytes) | (1 << offset)


def custom_hash(hash_class, *args, **kwargs):
    h = hash_class()
    for s in args:
        if s is not None:
            data = long_to_bytes(s) if isinstance(s, int) else s
            h.update(data)

    return bytes_to_long(h.digest())



import base64
import hashlib
from ssl import DER_cert_to_PEM_cert

import requests
from OpenSSL import crypto
from requests.adapters import HTTPAdapter
from urllib3.connectionpool import HTTPSConnectionPool
from urllib3.poolmanager import PoolManager
from urllib3.util.timeout import Timeout

from .constants import PUBKEY_HASH_DICT


class TLSPinningError(requests.exceptions.SSLError):
    def __init__(self, strerror):
        self.strerror = strerror
        super(TLSPinningError, self).__init__(strerror)


class TLSPinningHTTPSConnectionPool(HTTPSConnectionPool):
    """Verify the certificate upon each connection"""
    def __init__(
        self,
        host,
        hash_dict,
        port=None,
        strict=False,
        timeout=Timeout.DEFAULT_TIMEOUT,
        maxsize=1,
        block=False,
        headers=None,
        retries=None,
        _proxy=None,
        _proxy_headers=None,
        key_file=None,
        cert_file=None,
        cert_reqs=None,
        key_password=None,
        ca_certs=None,
        ssl_version=None,
        assert_hostname=None,
        assert_fingerprint=None,
        ca_cert_dir=None,
        **conn_kw
    ):
        self.hash_dict = hash_dict
        super(TLSPinningHTTPSConnectionPool, self).__init__(
            host,
            port,
            strict,
            timeout,
            maxsize,
            block,
            headers,
            retries,
            _proxy,
            _proxy_headers,
            key_file,
            cert_file,
            cert_reqs,
            key_password,
            ca_certs,
            ssl_version,
            assert_hostname,
            assert_fingerprint,
            ca_cert_dir,
            **conn_kw
        )

    def _validate_conn(self, conn):
        r = super(TLSPinningHTTPSConnectionPool, self)._validate_conn(conn)

        sock = conn.sock

        pem_certificate = self.get_certificate(sock)

        if self.is_session_secure(pem_certificate, conn, self.hash_dict):
            return r

    def is_session_secure(self, cert, conn, hash_dict):
        """Check if connection is secure"""

        cert_hash = self.extract_hash(cert)

        if not self.is_hash_valid(cert_hash, hash_dict):
            # Also generate a report
            conn.close()
            raise TLSPinningError("Insecure connection")

        return True

    def is_hash_valid(self, cert_hash, hash_dict):
        """Validate the hash against a known list of hashes/pins"""
        try:
            hash_dict[self.host].index(cert_hash)
        except (ValueError, KeyError, TypeError):
            return False
        else:
            return True

    def get_certificate(self, sock):
        """Extract and convert certificate to PEM format"""
        certificate_binary_form = sock.getpeercert(True)
        return DER_cert_to_PEM_cert(certificate_binary_form)

    def extract_hash(self, cert):
        """Extract encrypted hash from the certificate"""
        cert_obj = crypto.load_certificate(crypto.FILETYPE_PEM, cert)
        pubkey_obj = cert_obj.get_pubkey()
        pubkey = crypto.dump_publickey(crypto.FILETYPE_ASN1, pubkey_obj)

        spki_hash = hashlib.sha256(pubkey).digest()
        cert_hash = base64.b64encode(spki_hash).decode('utf-8')
        return cert_hash


class TLSPinningPoolManager(PoolManager):
    """Attach TLSPinningHTTPSConnectionPool to TLSPinningPoolManager"""
    def __init__(
        self,
        hash_dict,
        num_pools=10,
        headers=None,
        **connection_pool_kw
    ):
        self.hash_dict = hash_dict
        super(TLSPinningPoolManager, self).__init__(
            num_pools=10, headers=None, **connection_pool_kw
        )

    def _new_pool(self, scheme, host, port, request_context):
        if scheme != 'https':
            return super(TLSPinningPoolManager, self)._new_pool(
                scheme, host, port, request_context
                )

        kwargs = self.connection_pool_kw

        pool = TLSPinningHTTPSConnectionPool(
                    host=host, port=port,
                    hash_dict=self.hash_dict, **kwargs
                )

        return pool


class TLSPinningAdapter(HTTPAdapter):
    """Attach TLSPinningPoolManager to TLSPinningAdapter"""
    def __init__(self, hash_dict=PUBKEY_HASH_DICT):
        self.hash_dict = hash_dict
        super(TLSPinningAdapter, self).__init__()

    def init_poolmanager(
        self,
        connections,
        maxsize,
        block=False,
        **pool_kwargs
    ):
        self.poolmanager = TLSPinningPoolManager(
            num_pools=connections, maxsize=maxsize, block=block,
            strict=True, hash_dict=self.hash_dict, **pool_kwargs
        )



PUBKEY_HASH_DICT = {
    "api.protonvpn.ch": [
        "IEwk65VSaxv3s1/88vF/rM8PauJoIun3rzVCX5mLS3M=",
        "drtmcR2kFkM8qJClsuWgUzxgBkePfRCkRpqUesyDmeE=", 
        "YRGlaY0jyJ4Jw2/4M8FIftwbDIQfh8Sdro96CeEel54=", 
        "AfMENBVvOS8MnISprtvyPsjKlPooqh8nMB/pvCrpJpw="
    ],
    "protonvpn.com": [
        "+0dMG0qG2Ga+dNE8uktwMm7dv6RFEXwBoBjQ43GqsQ0=",
        "8joiNBdqaYiQpKskgtkJsqRxF7zN0C0aqfi8DacknnI=",
        "JMI8yrbc6jB1FYGyyWRLFTmDNgIszrNEMGlgy972e7w=",
        "Iu44zU84EOCZ9vx/vz67/MRVrxF1IO4i4NIa8ETwiIY="
    ]
    
}


from .srp import User as PmsrpUser

import requests, gnupg, base64
from .cert_pinning import TLSPinningAdapter

class ProtonError(Exception):
    def __init__(self, ret):
        super().__init__("[{0[Code]}] {0[Error]}".format(ret))
        self.code = ret['Code']
        self.error = ret['Error']

class Session:
    _base_headers = {
        "x-pm-apiversion": "3",
        "Accept": "application/vnd.protonmail.v1+json"
    }

    _srp_modulus_key = """-----BEGIN PGP PUBLIC KEY BLOCK-----
xjMEXAHLgxYJKwYBBAHaRw8BAQdAFurWXXwjTemqjD7CXjXVyKf0of7n9Ctm
L8v9enkzggHNEnByb3RvbkBzcnAubW9kdWx1c8J3BBAWCgApBQJcAcuDBgsJ
BwgDAgkQNQWFxOlRjyYEFQgKAgMWAgECGQECGwMCHgEAAPGRAP9sauJsW12U
MnTQUZpsbJb53d0Wv55mZIIiJL2XulpWPQD/V6NglBd96lZKBmInSXX/kXat
Sv+y0io+LR8i2+jV+AbOOARcAcuDEgorBgEEAZdVAQUBAQdAeJHUz1c9+KfE
kSIgcBRE3WuXC4oj5a2/U3oASExGDW4DAQgHwmEEGBYIABMFAlwBy4MJEDUF
hcTpUY8mAhsMAAD/XQD8DxNI6E78meodQI+wLsrKLeHn32iLvUqJbVDhfWSU
WO4BAMcm1u02t4VKw++ttECPt+HUgPUq5pqQWe5Q2cW4TMsE
=Y4Mw
-----END PGP PUBLIC KEY BLOCK-----"""

    @staticmethod
    def load(dump, TLSPinning=True):
        api_url = dump['api_url']
        appversion = dump['appversion']
        cookies = dump.get('cookies', {})
        s = Session(api_url, appversion, TLSPinning=TLSPinning)
        requests.utils.add_dict_to_cookiejar(s.s.cookies, cookies)
        s._session_data = dump['session_data']
        if s.UID is not None:
            s.s.headers['x-pm-uid'] = s.UID
            s.s.headers['Authorization'] = 'Bearer ' + s.AccessToken
        return s

    def dump(self):
        return {
            'api_url': self.__api_url,
            'appversion': self.__appversion,
            'cookies': self.s.cookies.get_dict(),
            'session_data': self._session_data
        }

    def __init__(self, api_url, appversion="Other", TLSPinning=True):
        self.__api_url = api_url
        self.__appversion = appversion

        ## Verify modulus
        self.__gnupg = gnupg.GPG()
        self.__gnupg.import_keys(self._srp_modulus_key)

        self._session_data = {}

        self.s = requests.Session()
        if TLSPinning:
            self.s.mount(self.__api_url, TLSPinningAdapter())
        self.s.headers['x-pm-appversion'] = appversion

    def api_request(self, endpoint, jsondata=None, additional_headers=None, method=None):
        fct = self.s.post
        if method is None:
            if jsondata is None:
                fct = self.s.get
            else:
                fct = self.s.post
        else:
            fct = {
                'get': self.s.get,
                'post': self.s.post,
                'put': self.s.put,
                'delete': self.s.delete,
                'patch': self.s.patch
            }.get(method.lower())

        if fct is None:
            raise ValueError("Unknown method: {}".format(method))

        ret = fct(
            self.__api_url + endpoint,
            headers = additional_headers,
            json = jsondata
        ).json()

        if ret['Code'] != 1000:
            raise ProtonError(ret)

        return ret


    def authenticate(self, username, password):
        self.logout()

        info_response = self.api_request("/auth/info", {"Username": username})
        d = self.__gnupg.decrypt(info_response['Modulus'])

        if not d.valid:
            raise ValueError('Invalid modulus')

        modulus   = base64.b64decode(d.data.strip())
        challenge = base64.b64decode(info_response["ServerEphemeral"])
        salt      = base64.b64decode(info_response["Salt"])
        session   = info_response["SRPSession"]
        version   = info_response["Version"]

        usr        = PmsrpUser(password, modulus)
        A          = usr.get_challenge()
        M          = usr.process_challenge(salt, challenge, version)

        if M is None:
            raise ValueError('Invalid challenge')

        ## Send response
        auth_response = self.api_request("/auth",
            {
                "Username": username,
                "ClientEphemeral" : base64.b64encode(A).decode('utf8'),
                "ClientProof" : base64.b64encode(M).decode('utf8'),
                "SRPSession": session,
            }
        )

        if "ServerProof" not in auth_response:
            raise ValueError("Invalid password")

        usr.verify_session( base64.b64decode(auth_response["ServerProof"]))
        if not usr.authenticated():
            raise ValueError('Invalid server proof')

        self._session_data = {
            'UID': auth_response["UID"],
            'AccessToken': auth_response["AccessToken"],
            'RefreshToken': auth_response["RefreshToken"],
            'Scope': auth_response["Scope"].split(),
        }

        if self.UID is not None:
            self.s.headers['x-pm-uid'] = self.UID
            self.s.headers['Authorization'] = 'Bearer ' + self.AccessToken

        return self.Scope

    def provide_2fa(self, code):
        ret = self.api_request('/auth/2fa', {
            "TwoFactorCode": code
        })
        self._session_data['Scope'] = ret['Scope']

        return self.Scope

    def logout(self):
        if self._session_data:
            self.api_request('/auth',method='DELETE')
            del self.s.headers['Authorization']
            del self.s.headers['x-pm-uid']
            self._session_data = {}

    def refresh(self):
        refresh_response = self.api_request('/auth/refresh', {
            "ResponseType": "token",
            "GrantType": "refresh_token",
            "RefreshToken": self.RefreshToken,
            "RedirectURI": "http://protonmail.ch"
        })
        self._session_data['AccessToken'] = refresh_response["AccessToken"]
        self._session_data['RefreshToken'] = refresh_response["RefreshToken"]
        self.s.headers['Authorization'] = 'Bearer ' + self.AccessToken

    @property
    def UID(self):
        return self._session_data.get('UID', None)

    @property
    def AccessToken(self):
        return self._session_data.get('AccessToken', None)

    @property
    def RefreshToken(self):
        return self._session_data.get('RefreshToken', None)

    @property
    def Scope(self):
        return self._session_data.get('Scope', [])
